@@ -27,20 +27,28 @@ void silc_say_error(char *msg, ...);
 void silc_channel_message(SilcClient client, SilcClientConnection conn,
 			  SilcClientEntry sender, 
 			  SilcChannelEntry channel, 
-			  SilcMessageFlags flags, char *msg);
+			  SilcMessagePayload payload,
+			  SilcMessageFlags flags, 
+			  const unsigned char *message,
+			  SilcUInt32 message_len);
 void silc_private_message(SilcClient client, SilcClientConnection conn,
 			  SilcClientEntry sender, 
-			  SilcMessageFlags flags, char *msg);
+			  SilcMessagePayload payload,
+			  SilcMessageFlags flags, 
+			  const unsigned char *message,
+			  SilcUInt32 message_len);
 void silc_notify(SilcClient client, SilcClientConnection conn, 
 		 SilcNotifyType type, ...);
 void silc_command(SilcClient client, SilcClientConnection conn, 
-		  SilcClientCommandContext cmd_context, int success,
-		  SilcCommand command);
+		  SilcClientCommandContext cmd_context, bool success,
+		  SilcCommand command, SilcStatus status);
 void silc_command_reply(SilcClient client, SilcClientConnection conn,
-			SilcCommandPayload cmd_payload, int success,
-			SilcCommand command, SilcCommandStatus status, ...);
-void silc_connect(SilcClient client, SilcClientConnection conn, int success);
-void silc_disconnect(SilcClient client, SilcClientConnection conn);
+			SilcCommandPayload cmd_payload, bool success,
+			SilcCommand command, SilcStatus status, ...);
+void silc_connect(SilcClient client, SilcClientConnection conn,
+		  SilcClientConnectionStatus status);
+void silc_disconnect(SilcClient client, SilcClientConnection conn,
+		     SilcStatus status, const char *message);
 void silc_ask_passphrase(SilcClient client, SilcClientConnection conn,
 			 SilcAskPassphrase completion, void *context);
 void silc_verify_public_key(SilcClient client, SilcClientConnection conn,
@@ -52,12 +60,19 @@ void silc_get_auth_method(SilcClient client, SilcClientConnection conn,
 			  SilcGetAuthMeth completion, void *context);
 void silc_failure(SilcClient client, SilcClientConnection conn, 
 		  SilcProtocol protocol, void *failure);
-int silc_key_agreement(SilcClient client, SilcClientConnection conn,
-		       SilcClientEntry client_entry, const char *hostname,
-		       SilcUInt16 port, SilcKeyAgreementCallback *completion,
-		       void **context);
+bool silc_key_agreement(SilcClient client, SilcClientConnection conn,
+		        SilcClientEntry client_entry, const char *hostname,
+		        SilcUInt16 port, SilcKeyAgreementCallback *completion,
+		        void **context);
 void silc_ftp(SilcClient client, SilcClientConnection conn,
 	      SilcClientEntry client_entry, SilcUInt32 session_id,
 	      const char *hostname, SilcUInt16 port);
+void
+silc_detach(SilcClient client, SilcClientConnection conn,
+            const unsigned char *detach_data, SilcUInt32 detach_data_len);
+char *
+silc_unescape_data(const char *escaped_data, SilcUInt32 *length);
+char * 
+silc_escape_data(const char *data, SilcUInt32 len);
 
 #endif
