@@ -17,14 +17,15 @@
   GNU General Public License for more details.
 
 */
-/* $Id: client_keyagr.c,v 1.22 2002/01/22 18:31:15 priikone Exp $ */
+/* $Id: client_keyagr.c,v 1.27 2002/10/20 16:57:55 priikone Exp $ */
 /* This file includes the Key Agreement packet processing and actual
    key agreement routines. This file has nothing to do with the actual
    connection key exchange protocol, it is implemented in the client.c
    and in protocol.c. This file implements the client-to-client key 
    agreement as defined by the SILC protocol. */
 
-#include "clientlibincludes.h"
+#include "silcincludes.h"
+#include "silcclient.h"
 #include "client_internal.h"
 
 SILC_TASK_CALLBACK(silc_client_key_agreement_final);
@@ -298,7 +299,7 @@ void silc_client_send_key_agreement(SilcClient client,
 				    const char *hostname,
 				    const char *bindhost,
 				    int port,
-				    uint32 timeout_secs,
+				    SilcUInt32 timeout_secs,
 				    SilcKeyAgreementCallback completion,
 				    void *context)
 {
@@ -306,9 +307,21 @@ void silc_client_send_key_agreement(SilcClient client,
   SilcClientKeyAgreement ke = NULL;
   SilcBuffer buffer;
 
-  if (!client_entry || client_entry->ke)
+  if (!client_entry)
     return;
 
+  if (client_entry->ke) {
+    completion(client, conn, client_entry, SILC_KEY_AGREEMENT_ALREADY_STARTED,
+	       NULL, context);
+    return;
+  }
+
+  if (client_entry == conn->local_entry) {
+    completion(client, conn, client_entry, SILC_KEY_AGREEMENT_SELF_DENIED,
+	       NULL, context);
+    return;
+  }
+
   /* Create the listener if hostname and port was provided.
    * also, use bindhost if it was specified.
    */
@@ -329,7 +342,6 @@ void silc_client_send_key_agreement(SilcClient client,
       completion(client, conn, client_entry, SILC_KEY_AGREEMENT_FAILURE,
 		 NULL, context);
       silc_free(ke);
-
       return;
     }
 
@@ -366,7 +378,6 @@ void silc_client_send_key_agreement(SilcClient client,
 			  client_entry->id, SILC_ID_CLIENT, NULL, NULL,
 			  buffer->data, buffer->len, FALSE);
   silc_buffer_free(buffer);
-
 }
 
 static int 
@@ -386,7 +397,8 @@ silc_client_connect_to_client_internal(SilcClientInternalConnectContext *ctx)
 				     (void *)ctx, 0, 0, 
 				     SILC_TASK_FD,
 				     SILC_TASK_PRI_NORMAL);
-  silc_schedule_set_listen_fd(ctx->client->schedule, sock, SILC_TASK_WRITE);
+  silc_schedule_set_listen_fd(ctx->client->schedule, sock, SILC_TASK_WRITE,
+			      FALSE);
 
   ctx->sock = sock;
 
@@ -516,9 +528,21 @@ void silc_client_perform_key_agreement(SilcClient client,
 
   SILC_LOG_DEBUG(("Start"));
 
-  if (!client_entry || !hostname || !port)
+  if (!client_entry)
     return;
 
+  if (!hostname || !port) {
+    completion(client, conn, client_entry, SILC_KEY_AGREEMENT_FAILURE,
+	       NULL, context);
+    return;
+  }
+
+  if (client_entry == conn->local_entry) {
+    completion(client, conn, client_entry, SILC_KEY_AGREEMENT_SELF_DENIED,
+	       NULL, context);
+    return;
+  }
+
   ke = silc_calloc(1, sizeof(*ke));
   ke->client = client;
   ke->conn = conn;
@@ -558,6 +582,12 @@ void silc_client_perform_key_agreement_fd(SilcClient client,
   if (!client_entry)
     return;
 
+  if (client_entry == conn->local_entry) {
+    completion(client, conn, client_entry, SILC_KEY_AGREEMENT_SELF_DENIED,
+	       NULL, context);
+    return;
+  }
+
   ke = silc_calloc(1, sizeof(*ke));
   ke->client = client;
   ke->conn = conn;
@@ -644,7 +674,7 @@ static void
 silc_client_key_agreement_resolve_cb(SilcClient client,
 				     SilcClientConnection conn,
 				     SilcClientEntry *clients,
-				     uint32 clients_count,
+				     SilcUInt32 clients_count,
 				     void *context)
 {
   SilcPacketContext *packet = (SilcPacketContext *)context;
@@ -704,6 +734,7 @@ void silc_client_key_agreement(SilcClient client,
     return;
 
   silc_client_get_client_by_id_resolve(client, sock->user_data, remote_id,
+				       NULL,
 				       silc_client_key_agreement_resolve_cb,
 				       silc_packet_context_dup(packet));
   silc_free(remote_id);
